/*
 * Copyright (c) 2004, Stef Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *     * Redistributions of source code must retain the above
 *       copyright notice, this list of conditions and the
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the
 *       above copyright notice, this list of conditions and
 *       the following disclaimer in the documentation and/or
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be
 *       used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *
 * CONTRIBUTORS
 *  Stef Walter <stef@memberwebs.com>
 *
 */

package com.memberwebs.httpauth;

import java.io.IOException;
import java.net.SocketException;
import java.util.Enumeration;

/**
 * Base class for HttpAuth authenticators. This class in not generally used
 * outside of the HttpAuth code.
 *
 * @author Stef Walter
 */
public abstract class BaseHttpAuthenticator
{
    protected HttpAuthConnectionSource m_source;

    protected BaseHttpAuthenticator(HttpAuthConnectionSource source)
    {
        m_source = source;
    }

    protected abstract void addResponseHeader(Object response, String name, String value);
    protected abstract void setResponseResult(Object response, int code)
        throws IOException;
    protected abstract Enumeration getHeaderValues(Object request, String name);
    protected abstract Enumeration getHeaderNames(Object request);

    protected String authenticateRequest(Object request, Object response,
                                         String connid, String method,
                                         String uri, String[] authtypes)
        throws HttpAuthException
    {
        // Prepare and send the request line
        String args = connid + " " + method + " " + uri;
        HttpAuthConnection conn = null;
        HttpAuthConnection.Response rsp = null;

        try
        {
            // Retry loop
            for(;;)
            {
                conn = m_source.getOpenConnection();

                try
                {
                    conn.sendLine("AUTH " + args);
                    sendHeaders(conn, request, authtypes);

                    rsp = conn.readResponse(200);
                    copyHeaders(conn, response, authtypes);

                    if(rsp.ccode != 200)
                    {
                        setResponseResult(response, rsp.ccode);
                        return null;
                    }
                    else
                    {
                        return rsp.details;
                    }
                }
                catch(SocketException e)
                {
                    // Always close connections that have errors on them
                    conn.close();

                    // Stupid hack but java doesn't let us do any better just now
                    if(!e.getMessage().toLowerCase().startsWith("broken pipe"))
                        throw new HttpAuthException("error communicating with httpauthd", e);
                }

                // The following code only gets executed when our sockets
                // have been disconnected. We loop up and try again. We're
                // pretty safe in doing this because the act of opening a
                // HttpAuthConnection sends some data through it, so only
                // stale connections should end up here.
                m_source.doneWithConnection(conn);
            }

        }
        catch(IOException e)
        {
            try
            {
                // Always close connections that have errors on them
                conn.close();
            }
            catch(IOException ex)
            	{ }

            throw new HttpAuthException("error communicating with httpauthd", e);
        }
        finally
        {
            if(conn != null)
                m_source.doneWithConnection(conn);
        }
    }

    protected void copyHeaders(HttpAuthConnection conn, Object response,
                               String[] authtypes)
        throws IOException
    {
        for(;;)
        {
            String line = conn.readLine();
            if(line != null)
            {
                line.trim();
                if(line.length() == 0)
                    line = null;
            }

            if(line == null)
                break;

            int pos = line.indexOf(":");

            // Skip invalid lines
            if(pos == -1)
                continue;

            String name = line.substring(0, pos).trim();
            String value = line.substring(pos + 1).trim();

            // Drop junk that's not allowed
            if(authtypes != null && authtypes.length > 0)
            {
                if(name.equalsIgnoreCase("WWW-Authenticate"))
                {
                    boolean ok = false;

                    for(int i = 0; i < authtypes.length; i++)
                    {
                        if(value.startsWith(authtypes[i]))
                        {
                            ok = true;
                            break;
                        }
                    }

                    if(!ok)
                        continue;
                }
            }

            addResponseHeader(response, name, value);
        }
    }

    /**
     * @param req
     */
    protected void sendHeaders(HttpAuthConnection conn, Object request,
                               String[] authtypes)
        throws IOException
    {
        Enumeration e = getHeaderNames(request);

        while(e.hasMoreElements())
        {
            String name = (String)e.nextElement();
            if(!name.equalsIgnoreCase("Authorization"))
                continue;

            Enumeration e2 = getHeaderValues(request, name);
            while(e2.hasMoreElements())
            {
                String value = (String)e2.nextElement();

                if(authtypes == null || authtypes.length == 0)
                {
                    conn.sendLine(name + ": " + value);
                }
                else
                {
                    for(int i = 0; i < authtypes.length; i++)
                    {
                        if(value.startsWith(authtypes[i]))
                        {
                            conn.sendLine(name + ": " + value);
                            break;
                        }
                    }
                }
            }
        }

        // Headers always end in a blank line
        conn.sendLine("");
    }
}
