/*
 * Copyright (c) 2004, Stefan Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *     * Redistributions of source code must retain the above
 *       copyright notice, this list of conditions and the
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the
 *       above copyright notice, this list of conditions and
 *       the following disclaimer in the documentation and/or
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be
 *       used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *
 * CONTRIBUTORS
 *  Stef Walter <stef@memberwebs.com>
 *
 */

package com.memberwebs.httpauth.servlet;

import java.io.IOException;
import java.util.Enumeration;

import javax.servlet.ServletException;
import javax.servlet.UnavailableException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.memberwebs.httpauth.BaseHttpAuthenticator;
import com.memberwebs.httpauth.HttpAuthConnectionSource;
import com.memberwebs.httpauth.HttpAuthException;

/**
 * A class which can be used by servlets to authenticate the user for
 * a given request.
 *
 * Note that the user will not be set on the request and will not be
 * available via req.getRemoteUser().
 *
 * @author Stef Walter
 */

public class ServletHttpAuthenticator
    extends BaseHttpAuthenticator
{
    /**
     * Creates a new ServletHttpAuthenticator.
     *
     * @param source A source of HttpAuth connections
     *               (such as {@link #HttpAuthConnectionPool})
     */
    public ServletHttpAuthenticator(HttpAuthConnectionSource source)
    {
        super(source);
    }

    /**
     * Authenticate a request. If the authentication fails, headers and
     * a response will be sent.
     *
     * @param req The HTTP request.
     * @param resp The HTTP response used when authentication fails.
     * @return The authenticated user name, or null if unsuccessful.
     */
    public String authenticate(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException
    {
        return authenticate(req, resp, null);
    }

    /**
     * Authenticate a request. If the authentication fails, headers and
     * a response will be sent. The caller can limit the types of
     * authentication allowed.
     *
     * @param req The HTTP request.
     * @param resp The HTTP response used when authentication fails.
     * @param authtypes The types of authentication allowed.
     * @return The authenticated user name, or null if unsuccessful.
     */
    public String authenticate(HttpServletRequest req, HttpServletResponse resp,
                               String[] authtypes)
        throws ServletException
    {
        String method = req.getMethod();
        String uri = req.getRequestURI();

        String qs = req.getQueryString();
        if(qs != null)
            uri += "?" + qs;

        // NOTE: We can't set the remote user on the request. That's just
        // the way the interfaces are designed

        try
        {
            // Connection ID is random. This prevents NTLM from working :(
            String connid = "" + Math.random();
            return authenticateRequest(req, resp, connid, method, uri, authtypes);
        }
        catch(HttpAuthException e)
        {
            throw new UnavailableException(e.getMessage(), 30);
        }
    }

    /* -------------------------------------------------------------------------
     * OVERRIDES
     */

    protected void addResponseHeader(Object response, String name, String value)
    {
        HttpServletResponse resp = (HttpServletResponse)response;
        resp.addHeader(name, value);
    }

    protected void setResponseResult(Object response, int code)
        throws IOException
    {
        HttpServletResponse resp = (HttpServletResponse)response;
        if(code >= 200 && code < 300)
            resp.setStatus(code);
        else
            resp.sendError(code);
    }

    protected Enumeration getHeaderValues(Object request, String name)
    {
        HttpServletRequest req = (HttpServletRequest)request;
        return req.getHeaders(name);
    }

    protected Enumeration getHeaderNames(Object request)
    {
        HttpServletRequest req = (HttpServletRequest)request;
        return req.getHeaderNames();
    }
}
